//////////////////////////////////////////////////////////////////////
// JetSQL - MS Jet SQL script interpreter (c) TECHSYS HW a SW s.r.o.
//////////////////////////////////////////////////////////////////////
// Module: Interpreter.cpp
// Author: Pavel Klocek
// Desc:   Implementation class CSqlInterpreter 
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "LexicalAnalyser.h"
#include "Interpreter.h"
#include "Preparser.h"

using namespace std;

//size of line buffer
#define MAXLINELN 1000

extern char * stristr( char * pszSource, const char * pcszSearch );

//@func CString | VarToStr | Konverze COleVariant na CString
CString VarToStr(COleVariant & v)
{
	if(v.vt == VT_BSTR) return (LPCTSTR)v.bstrVal;
	if(v.vt == VT_UI1)
	{
		char buf[10];
		sprintf(buf,"%i",v.iVal);
		return buf;
	}
	if(v.vt == VT_I2)
	{
		char buf[50];
		sprintf(buf,"%i",v.iVal);
		return buf;
	}
	if(v.vt == VT_I4)
	{
		char buf[50];
		sprintf(buf,"%i",v.lVal);
		return buf;
	}
	if(v.vt == VT_R4)
	{
		char buf[50];
		sprintf(buf,"%g",v.fltVal);
		return buf;
	}
	if(v.vt == VT_R8)
	{
		char buf[50];
		sprintf(buf,"%g",v.dblVal);
		return buf;
	}
	if(v.vt == VT_BOOL)
	{
		if(v.boolVal==0) return "FALSE";
		else return "TRUE";
	}
	if(v.vt == VT_DATE)
	{
		COleDateTime dt(v.date);
		return dt.Format();
	}
	if(v.vt == VT_CY)
	{
		COleCurrency cy(v.cyVal);
		return cy.Format();
	}
	if(v.vt == VT_NULL) return "";
	if(v.vt == VT_EMPTY) return "nothing";
	if(v.vt == 8209) return "<<binary>>";
	return "?format";
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSqlInterpreter::CSqlInterpreter()
{
//*** initialize lexical analyser for SQL
	m_lexan.setSymbol(CSymbol(",",commaT));
	m_lexan.setSymbol(CSymbol("(",leftpT));
	m_lexan.setSymbol(CSymbol(")",rightpT));
	m_lexan.setSymbol(CSymbol("=",ekvT));
	m_lexan.setSymbol(CSymbol(".",dotT));
	m_lexan.setSymbol(CSymbol("*",asterixT));
	//data types
	m_lexan.setSymbol(CSymbol("ALPHANUMERIC",textT));
	m_lexan.setSymbol(CSymbol("AUTOINCREMENT",counterT));
	m_lexan.setSymbol(CSymbol("BINARY",binaryT));
	m_lexan.setSymbol(CSymbol("BIT",bitT));
	m_lexan.setSymbol(CSymbol("BOOLEAN",bitT));
	m_lexan.setSymbol(CSymbol("BYTE",byteT));
	m_lexan.setSymbol(CSymbol("CHAR",textT));
	m_lexan.setSymbol(CSymbol("CHARACTER",textT));
	m_lexan.setSymbol(CSymbol("COUNTER",counterT));
	m_lexan.setSymbol(CSymbol("CURRENCY",currencyT));
	m_lexan.setSymbol(CSymbol("DATE",datetimeT));
	m_lexan.setSymbol(CSymbol("DATETIME",datetimeT));
	m_lexan.setSymbol(CSymbol("DOUBLE",doubleT));
	m_lexan.setSymbol(CSymbol("FLOAT",doubleT));
	m_lexan.setSymbol(CSymbol("FLOAT8",doubleT));
	m_lexan.setSymbol(CSymbol("FLOAT4",singleT));
	m_lexan.setSymbol(CSymbol("GENERAL",longbinaryT));
	m_lexan.setSymbol(CSymbol("GUID",guidT));
	m_lexan.setSymbol(CSymbol("SMALLINT",shortT));
	m_lexan.setSymbol(CSymbol("IEEEDOUBLE",doubleT));
	m_lexan.setSymbol(CSymbol("IEEESINGLE",singleT));
	m_lexan.setSymbol(CSymbol("INT",longT));
	m_lexan.setSymbol(CSymbol("INTEGER",longT));
	m_lexan.setSymbol(CSymbol("INTEGER4",longT));
	m_lexan.setSymbol(CSymbol("INTEGER1",byteT));
	m_lexan.setSymbol(CSymbol("INTEGER2",shortT));
	m_lexan.setSymbol(CSymbol("LOGICAL",bitT));
	m_lexan.setSymbol(CSymbol("LOGICAL1",bitT));
	m_lexan.setSymbol(CSymbol("LONG",longT));
	m_lexan.setSymbol(CSymbol("LONGBINARY",longbinaryT));
	m_lexan.setSymbol(CSymbol("LONGTEXT",longtextT));
	m_lexan.setSymbol(CSymbol("MEMO",longtextT));
	m_lexan.setSymbol(CSymbol("MONEY",currencyT));
	m_lexan.setSymbol(CSymbol("NUMBER",doubleT));
	m_lexan.setSymbol(CSymbol("NUMERIC",doubleT));
	m_lexan.setSymbol(CSymbol("OLEOBJECT",longbinaryT));
	m_lexan.setSymbol(CSymbol("REAL",singleT));
	m_lexan.setSymbol(CSymbol("SHORT",shortT));
	m_lexan.setSymbol(CSymbol("SINGLE",singleT));
	m_lexan.setSymbol(CSymbol("SMALLINT",shortT));
	m_lexan.setSymbol(CSymbol("STRING",textT));
	m_lexan.setSymbol(CSymbol("TEXT",textT));
	m_lexan.setSymbol(CSymbol("TIME",datetimeT));
	m_lexan.setSymbol(CSymbol("TIMESTAMP",datetimeT));
	m_lexan.setSymbol(CSymbol("VALUE",valueT));
	m_lexan.setSymbol(CSymbol("VARBINARY",binaryT));
	m_lexan.setSymbol(CSymbol("VARCHAR",textT));
	m_lexan.setSymbol(CSymbol("YESNO",bitT));
	//other reserved words
	m_lexan.setSymbol(CSymbol("ADD",addT));
	m_lexan.setSymbol(CSymbol("ALL",allT));
	m_lexan.setSymbol(CSymbol("ALTER",alterT));
	m_lexan.setSymbol(CSymbol("AND",andT));
	m_lexan.setSymbol(CSymbol("ANY",anyT));
	m_lexan.setSymbol(CSymbol("AS",asT));
	m_lexan.setSymbol(CSymbol("ASC",ascT));
	m_lexan.setSymbol(CSymbol("AVG",avgT));
	m_lexan.setSymbol(CSymbol("BETWEEN",betweenT));
	m_lexan.setSymbol(CSymbol("BY",byT));
	m_lexan.setSymbol(CSymbol("COLUMN",columnT));
	m_lexan.setSymbol(CSymbol("CONSTRAINT",constraintT));
	m_lexan.setSymbol(CSymbol("COUNT",countT));
	m_lexan.setSymbol(CSymbol("CREATE",createT));
	m_lexan.setSymbol(CSymbol("DATABASE",databaseT));
	m_lexan.setSymbol(CSymbol("DELETE",deleteT));
	m_lexan.setSymbol(CSymbol("DESC",descT));
	m_lexan.setSymbol(CSymbol("DISALLOW",disallowT));
	m_lexan.setSymbol(CSymbol("DISTINCT",distinctT));
	m_lexan.setSymbol(CSymbol("DISTINCTROW",distinctrowT));
	m_lexan.setSymbol(CSymbol("DROP",dropT));
	m_lexan.setSymbol(CSymbol("EQV",eqvT));
	m_lexan.setSymbol(CSymbol("EXISTS",existT));
	m_lexan.setSymbol(CSymbol("FOREIGN",foreignT));
	m_lexan.setSymbol(CSymbol("FROM",fromT));
	m_lexan.setSymbol(CSymbol("GROUP",groupT));
	m_lexan.setSymbol(CSymbol("HAVING",havingT));
	m_lexan.setSymbol(CSymbol("IGNORE",ignoreT));
	m_lexan.setSymbol(CSymbol("IMP",impT));
	m_lexan.setSymbol(CSymbol("IN",inT));
	m_lexan.setSymbol(CSymbol("INDEX",indexT));
	m_lexan.setSymbol(CSymbol("INNER",innerT));
	m_lexan.setSymbol(CSymbol("INSERT",insertT));
	m_lexan.setSymbol(CSymbol("INTO",intoT));
	m_lexan.setSymbol(CSymbol("IS",isT));
	m_lexan.setSymbol(CSymbol("JOIN",joinT));
	m_lexan.setSymbol(CSymbol("KEY",keyT));
	m_lexan.setSymbol(CSymbol("LEFT",leftT));
	m_lexan.setSymbol(CSymbol("LIKE",likeT));
	m_lexan.setSymbol(CSymbol("MAX",maxT));
	m_lexan.setSymbol(CSymbol("MIN",minT));
	m_lexan.setSymbol(CSymbol("MOD",modT));
	m_lexan.setSymbol(CSymbol("NOT",notT));
	m_lexan.setSymbol(CSymbol("NULL",nullT));
	m_lexan.setSymbol(CSymbol("ON",onT));
	m_lexan.setSymbol(CSymbol("OPTION",optionT));
	m_lexan.setSymbol(CSymbol("OR",orT));
	m_lexan.setSymbol(CSymbol("ORDER",orderT));
	m_lexan.setSymbol(CSymbol("OWNERACCESS",owneraccessT));
	m_lexan.setSymbol(CSymbol("PARAMETERS",parametersT));
	m_lexan.setSymbol(CSymbol("PERCENT",percentT));
	m_lexan.setSymbol(CSymbol("PIVOT",pivotT));
	m_lexan.setSymbol(CSymbol("PRIMARY",primaryT));
	m_lexan.setSymbol(CSymbol("PROCEDURE",procedureT));
	m_lexan.setSymbol(CSymbol("REFERENCES",referencesT));
	m_lexan.setSymbol(CSymbol("RIGHT",rightT));
	m_lexan.setSymbol(CSymbol("SELECT",selectT));
	m_lexan.setSymbol(CSymbol("SET",setT));
	m_lexan.setSymbol(CSymbol("SOME",someT));
	m_lexan.setSymbol(CSymbol("STDEV",stdevT));
	m_lexan.setSymbol(CSymbol("STDEVP",stdevpT));
	m_lexan.setSymbol(CSymbol("SUM",sumT));
	m_lexan.setSymbol(CSymbol("TABLE",tableT));
	m_lexan.setSymbol(CSymbol("TOP",topT));
	m_lexan.setSymbol(CSymbol("TRANSFORM",transformT));
	m_lexan.setSymbol(CSymbol("UNION",unionT));
	m_lexan.setSymbol(CSymbol("UNIQUE",uniqueT));
	m_lexan.setSymbol(CSymbol("UPDATE",updateT));
	m_lexan.setSymbol(CSymbol("VALUES",valuesT));
	m_lexan.setSymbol(CSymbol("VAR",varT));
	m_lexan.setSymbol(CSymbol("VARP",varpT));
	m_lexan.setSymbol(CSymbol("VIEW",viewT));
	m_lexan.setSymbol(CSymbol("WHERE",whereT));
	m_lexan.setSymbol(CSymbol("WITH",withT));
	m_lexan.setSymbol(CSymbol("XOR",xorT));


}

CSqlInterpreter::~CSqlInterpreter()
{
	
}

int CSqlInterpreter::OpenDatabase(CString dbfile, bool create)
{
	m_dbfile = dbfile;
	try
	{
		if(!create)
		{
			//opening database
			m_dbs.Open(dbfile);
		}
		else
		{
			//creating database
			if(remove(dbfile))
			{
				if(errno!=2)
				{
					CString err;
					err.Format("Can't delete old database %s!",dbfile);
					m_show(0,err,fatalerror);
					return 1;
				}
			}
			m_dbs.Create(dbfile);
		}
	}
	catch (CDaoException* e)
	{
		m_show(0,e->m_pErrorInfo->m_strDescription,fatalerror);
		e->Delete();
		return 1;
	}
	catch (...)
	{
		CString err;
		err.Format("Can't open database file %s!",dbfile);
		m_show(0,err,fatalerror);
		return 1;
	}
	return 0;
}

int CSqlInterpreter::ReopenDatabase()
{
	try
	{
		m_dbs.Close();
		m_dbs.Open(m_dbfile);
	}
	catch (CDaoException* e)
	{
		CString err;
		err.Format("Can't reopen the database: %s",e->m_pErrorInfo->m_strDescription);
		m_show(0,err,fatalerror);
		e->Delete();
		return 1;
	}
	catch (...)
	{
		m_show(0,"Can't reopen the database!",fatalerror);
		return 1;
	}
	return 0;
}

int CSqlInterpreter::Go(CString dbfile, CString sqlfile, SQLSHOWFUNC& show, bool create, bool verbose, bool echo)
{
	m_verbose = verbose;
	m_show = show;
	int errcnt = 0; //no errors at the beginning
	m_ln = 0;	  //we will start with the first line
	char line[MAXLINELN];
	CString pms;

	CPreparser p;

	//open database file
	if(OpenDatabase(dbfile,create)) return -1;

	//opening the stream
	ifstream in(sqlfile);
	if(!in.is_open())
	{
		pms.Format("Can't open SQL file %s!",sqlfile);
		m_show(0,pms,fatalerror);
		return -1;
	}

	while(in.getline(line, MAXLINELN)) 
	{
		m_ln++;
		
		//expand macros
		ExpandMacros(line,MAXLINELN);

		if(echo) m_show(m_ln,line,echoln);

		//single line comment handle immediately
		{
			char* t = line;
			while(isspace(*t))t++;
			if((t-line)<MAXLINELN-1 && t[0]=='-' && t[1]=='-') continue;
		}

		for(int ix=0; line[ix]; ix++)
		{
			try{
				if(p.put(line[ix]))
				{
					//command completed
					m_cmd = p.GetCommand();
					//lexical parsing
					m_lexan.Text2TokenSequence(m_cmd);
					//syntax checking and executing
					if(Execute()) errcnt++;
				}
			}
			catch(CString& e)
			{
				m_show(m_ln,e,error);
			}
			catch(...){}
		}
		if(p.m_state != CPreparser::start) p.put(' '); //for newline

	}

	return errcnt;
}

int CSqlInterpreter::Execute()
{
		CToken t = m_lexan.GetCurrentToken();
		CString pms;

		switch(t.getTokenType())
		{
		case selectT: 
			return Select();
		case createT: 
			m_lexan.NextToken();
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()==nilT) 
			{
				m_show(m_ln,"Invalid CREATE statement",error);
				return -1;
			}
			switch(t.getTokenType())
			{
			case tableT: return CreateTable();
			case indexT: return CreateIndex();
			case uniqueT:
				{
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
					if(t.getTokenType()!=indexT) 
					{
						m_show(m_ln,"Missing INDEX after CREATE UNIQUE",error);
						return -1;
					}
					return CreateIndex();
				}
			case viewT: return CreateView();

			}
			break;
		case insertT: return InsertInto();
		case deleteT: return Delete();
		case dropT: return Drop();
		case updateT: return Update();
		default:
			m_show(m_ln,"Unknown statement",error);
			return -1;
		}

	return 0;
}

int CSqlInterpreter::CreateTable()
{
	CString pms;
	//name of the table
	CToken t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=idT)
	{
		pms.Format("Invalid table name - %s",t.getTokenName());
		m_show(m_ln,pms,error);
		return -1;
	}

	//check existence
	if(TableExists(t.getTokenName()))
	{
		pms.Format("Table %s already exists",t.getTokenName());
		m_show(m_ln,pms,error);
		return -1;
	}
	
	CString table = t.getTokenName();

	//left parenthesis must follow
	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=leftpT)
	{
		pms.Format("Left parenthesis missing after name for table %s",t.getTokenName());
		m_show(m_ln,pms,error);
		return -1;
	}

	//iterating fields
	for(;;)
	{
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		
		
		if(t.getTokenType()==parametersT)
		{
			t.TokenType = idT;
			t.TokenName = "Parameters";
		}

		if(t.getTokenType()==constraintT) break;
		if(t.getTokenType()!=idT)
		{
			m_show(m_ln,"Field name or CONSTRAINT missing",error);
			return -1;
		}

		CString field = t.getTokenName();
	
		//field type
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()<binaryT && t.getTokenType()>valueT)
		{
			pms.Format("Field type missing for field %s",field);
			m_show(m_ln,pms,error);
			return -1;
		}
		
		//length must follow for text fields
		if(t.getTokenType()==textT || t.getTokenType()==binaryT)
		{
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=leftpT)
			{
				pms.Format("Field length missing for field %s",field);
				m_show(m_ln,pms,error);
				return -1;
			}
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=intT)
			{
				pms.Format("Field length missing for field %s",field);
				m_show(m_ln,pms,error);
				return -1;
			}
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=rightpT)
			{
				pms.Format("Right parenthesis missing after length of field %s",field);
				m_show(m_ln,pms,error);
				return -1;
			}
		}

		//comma of closing parenthesis
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()==notT)
		{
			//NOT NULL
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=nullT)
			{
				pms.Format("Solitary NOT after field %s definition",field);
				m_show(m_ln,pms,error);
				return -1;
			}
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();

		}
		if(t.getTokenType()==constraintT)
		{
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=idT)
			{
				pms.Format("Missing contraint name for field %s",field);
				m_show(m_ln,pms,error);
				return -1;
			}
			//constraint type
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=primaryT && t.getTokenType()!=uniqueT && t.getTokenType()!=notT && t.getTokenType()!=referencesT)
			{
				pms.Format("Missing contraint type for field %s",field);
				m_show(m_ln,pms,error);
				return -1;
			}
			if(t.getTokenType()==notT)
			{
				//NOT NULL
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
				if(t.getTokenType()!=nullT)
				{
					pms.Format("Missing NULL after NOT in constraint of field %s",field);
					m_show(m_ln,pms,error);
					return -1;
				}
			}
			if(t.getTokenType()==primaryT)
			{
				//NOT NULL
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
				if(t.getTokenType()!=keyT)
				{
					pms.Format("Missing KEY after PRIMARY in constraint of field %s",field);
					m_show(m_ln,pms,error);
					return -1;
				}
			}
			if(t.getTokenType()==referencesT)
			{
				//foreign table
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
				if(t.getTokenType()!=idT)
				{
					pms.Format("Missing name of referenced table in constraint of field %s",field);
					m_show(m_ln,pms,error);
					return -1;
				}
				if(!TableExists(t.getTokenName()))
				{
					pms.Format("Table %s referenced in constraint of field %s does not exist",t.getTokenName(),field);
					m_show(m_ln,pms,error);
					return -1;
				}
				//foreign field
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
				if(t.getTokenType()==leftpT)
				{
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
					if(t.getTokenType()!=idT)
					{
						pms.Format("Missing foreign field in constraint of field %s",field);
						m_show(m_ln,pms,error);
						return -1;
					}
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
					if(t.getTokenType()!=rightpT)
					{
						pms.Format("Missing right parenthesis after foreign field in constraint of field %s",field);
						m_show(m_ln,pms,error);
						return -1;
					}
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
				}
			}
			else
			{
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
			}
		}
		if(t.getTokenType()==rightpT) break;
		if(t.getTokenType()!=commaT)
		{
			pms.Format("Comma or right parenthesis missing after field %s definition",field);
			m_show(m_ln,pms,error);
			return -1;
		}
	}
	//CONSTRAINT clause
	if(t.getTokenType()==constraintT)
	{
		for(;;)
		{
			//name
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=idT && t.getTokenType()!=constraintT)
			{
				m_show(m_ln,"Constraint name missing",error);
				return -1;
			}
			//type
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=primaryT && t.getTokenType()!=uniqueT && t.getTokenType()!=notT && t.getTokenType()!=foreignT)
			{
				m_show(m_ln,"Missing contraint type",error);
				return -1;
			}
			if(t.getTokenType()==foreignT)
			{
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
				if(t.getTokenType()!=keyT)
				{
					m_show(m_ln,"KEY missing in constraint FOREIGN KEY",error);
					return -1;
				}
				//field list
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
				if(t.getTokenType()!=leftpT)
				{
					m_show(m_ln,"Field list missing in constraint",error);
					return -1;
				}
				for(;;)
				{
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
					if(t.getTokenType()!=idT)
					{
						m_show(m_ln,"Field name missing in constraint",error);
						return -1;
					}
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
					if(t.getTokenType()==rightpT) break;
					if(t.getTokenType()!=commaT)
					{
						m_show(m_ln,"Comma or right parenthesis missing in the field list of constraint",error);
						return -1;
					}
				}
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
				if(t.getTokenType()!=referencesT)
				{
					m_show(m_ln,"Missing REFERENCES in FOREIGN KEY INDEX constraint",error);
					return -1;
				}
				//foreign table
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
				if(t.getTokenType()!=idT)
				{
					m_show(m_ln,"Missing name of referenced table in constraint",error);
					return -1;
				}
				if(!TableExists(t.getTokenName()))
				{
					pms.Format("Table %s referenced in constraint does not exist",t.getTokenName());
					m_show(m_ln,pms,error);
					return -1;
				}
				t = m_lexan.GetCurrentToken();
				if(t.getTokenType()==leftpT)
				{
					for(;;)
					{
						t = m_lexan.GetCurrentToken();
						m_lexan.NextToken();
						if(t.getTokenType()!=idT)
						{
							m_show(m_ln,"Field name missing in constraint",error);
							return -1;
						}
						t = m_lexan.GetCurrentToken();
						m_lexan.NextToken();
						if(t.getTokenType()==rightpT) break;
						if(t.getTokenType()!=commaT)
						{
							m_show(m_ln,"Comma or right parenthesis missing in the field list of constraint",error);
							return -1;
						}
					}
				}
				
			}
			else
			{
				if(t.getTokenType()==primaryT)
				{
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
					if(t.getTokenType()!=keyT)
					{
						m_show(m_ln,"KEY missing in constraint PRIMARY KEY",error);
						return -1;
					}
				}
				if(t.getTokenType()==notT)
				{
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
					if(t.getTokenType()!=nullT)
					{
						m_show(m_ln,"NULL missing in constraint NOT NULL",error);
						return -1;
					}
				}
				//field list
				t = m_lexan.GetCurrentToken();
				m_lexan.NextToken();
				if(t.getTokenType()!=leftpT)
				{
					m_show(m_ln,"Field list missing in constraint",error);
					return -1;
				}
				for(;;)
				{
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
					if(t.getTokenType()!=idT)
					{
						m_show(m_ln,"Field name missing in constraint",error);
						return -1;
					}
					t = m_lexan.GetCurrentToken();
					m_lexan.NextToken();
					if(t.getTokenType()==rightpT) break;
					if(t.getTokenType()!=commaT)
					{
						m_show(m_ln,"Comma or right parenthesis missing in the field list of constraint",error);
						return -1;
					}
				}
			}
			if(t.getTokenType()==rightpT) break;
			if(t.getTokenType()!=commaT)
			{
				m_show(m_ln,"Comma or right parenthesis missing in the field list of constraint",error);
				return -1;
			}
		}
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=rightpT)
		{
			m_show(m_ln,"Right parenthesis missing",error);
			return -1;
		}
		
	}
	//check for following invalid text
	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=nilT && t.getTokenType()!=endfileT)
	{
		m_show(m_ln,"Semicolon missing behind the statement CREATE TABLE",error);
		return -1;
	}

	try
	{
		m_dbs.Execute(m_cmd);
	}
	catch (CDaoException* e)
	{
		m_show(m_ln,e->m_pErrorInfo->m_strDescription,error);
		e->Delete();
		return -1;
	}
	catch(...)
	{
		pms.Format("Error creating table %s",table);
		m_show(m_ln, pms, error);
		return -1;
	}

	ReopenDatabase();
	pms.Format("Table %s created",table);
	if(m_verbose) m_show(m_ln,pms,info);
	return 0;
}

bool CSqlInterpreter::TableExists(const char * name)
{
	CDaoTableDef td(&m_dbs);
	try
	{
		td.Open(name);
		return true;
	}
	catch(CDaoException* e){e->Delete();}
	catch(...){}
	return false;
}

bool CSqlInterpreter::IndexExists(const char * name)
{
	CDaoRelationInfo rel;
	try
	{
		m_dbs.GetRelationInfo(name,rel);
		return true;
	}
	catch(CDaoException* e){e->Delete();}
	catch(...){}
	return false;
}

bool CSqlInterpreter::QueryExists(const char * name)
{
	CDaoQueryDef qd(&m_dbs);
	try
	{
		qd.Open(name);
		return true;
	}
	catch(CDaoException* e){e->Delete();}
	catch(...){}
	return false;
}

int CSqlInterpreter::CreateIndex()
{
	//name of the index
	CToken t = m_lexan.GetCurrentToken();
	CString pms;
	m_lexan.NextToken();
	if(t.getTokenType()!=idT)
	{
		m_show(m_ln,"Invalid index name",error);
		return -1;
	}

	if(IndexExists(t.getTokenName()))
	{
		pms.Format("Index %s already exists",t.getTokenName());
		m_show(m_ln,pms,error);
		return -1;
	}
	
	CString index = t.getTokenName();

	//ON and name of a table must follow
	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=onT)
	{
		m_show(m_ln,"ON missing after index name",error);
		return -1;
	}
	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=idT)
	{
		pms.Format("Missing name of table in index %s",index);
		m_show(m_ln,pms,error);
		return -1;
	}
	if(!TableExists(t.getTokenName()))
	{
		pms.Format("Table %s in index %s does not exist",t.getTokenName(),index);
		m_show(m_ln,pms,error);
		return -1;
	}
	//field list
	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=leftpT)
	{
		m_show(m_ln,"Left parenthesis missing after table name",error);
		return -1;
	}
	for(;;)
	{
		//field name
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=idT)
		{
			m_show(m_ln,"Field name missing",error);
			return -1;
		}
		CString field = t.getTokenName();
		t = m_lexan.GetCurrentToken();
		if(t.getTokenType()==ascT || t.getTokenType()==descT)
		{
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
		}
		if(t.getTokenType()==rightpT) break;
		if(t.getTokenType()!=commaT)
		{
			pms.Format("Error in field list of index %s behind field %s",index,field);
			m_show(m_ln,pms,error);
			return -1;
		}
		m_lexan.NextToken();
	}
	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()== withT)
	{
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=primaryT && t.getTokenType()!=disallowT && t.getTokenType()!=ignoreT)
		{
			pms.Format("Incomplete WITH clause in index %s",index);
			m_show(m_ln,pms,error);
			return -1;
		}
		if(t.getTokenType()!=primaryT)
		{
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=nullT)
			{
				pms.Format("*** Incomplete WITH clause in index %s",index);
				m_show(m_ln,pms,error);
				return -1;
			}
		}
	}

	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=nilT && t.getTokenType()!=endfileT)
	{
		m_show(m_ln,"Semicolon missing behind the statement CREATE TABLE",error);
		return -1;
	}

	try
	{
		m_dbs.Execute(m_cmd);
	}
	catch (CDaoException* e)
	{
		m_show(m_ln,e->m_pErrorInfo->m_strDescription,error);
		e->Delete();
		return -1;
	}
	catch(...)
	{
		pms.Format("Error creating index %s",index);
		m_show(m_ln,pms,error);
		return -1;
	}

	ReopenDatabase();
	if(m_verbose)
	{
		pms.Format("Index %s created",index);
		m_show(m_ln,pms,info);
	}
	return 0;
}

int CSqlInterpreter::InsertInto()
{
	CString pms;
	m_lexan.NextToken();
	CToken t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=intoT)
	{
		m_show(m_ln,"Missing INTO in INSERT INTO statement",error);
		return -1;
	}
	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=idT)
	{
		m_show(m_ln,"Table name missing in INSERT INTO statement",info);
		return -1;
	}
	if(!TableExists(t.getTokenName()))
	{
		pms.Format("Table %s does not exists",t.getTokenName());
		m_show(m_ln,pms,info);
		return -1;
	}
	
	CString table = t.getTokenName();

	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()==leftpT)
	{
		for(;;)
		{
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()!=idT)
			{
				m_show(m_ln,"Field name missing",info);
				return -1;
			}
			t = m_lexan.GetCurrentToken();
			m_lexan.NextToken();
			if(t.getTokenType()==rightpT) break;
			if(t.getTokenType()!=commaT)
			{
				m_show(m_ln,"Comma or right parenthesis missing in the field list of constraint",info);
				return -1;
			}
		}
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
	}

	int ra = 1;
	try
	{
		m_dbs.Execute(m_cmd);
		ra = m_dbs.GetRecordsAffected();
	}
	catch (CDaoException* e)
	{
		m_show(m_ln,e->m_pErrorInfo->m_strDescription,error);
		e->Delete();
		return -1;
	}
	catch(...)
	{
		pms.Format("Error inserting into %s",table);
		m_show(m_ln,pms,error);
		return -1;
	}

	if(m_verbose) 
	{
		pms.Format("%i record(s) inserted into table %s",ra,table);
		m_show(m_ln,pms,info);
	}

	return 0;
}

int CSqlInterpreter::Delete()
{
	CString pms;
	m_lexan.NextToken();
	CToken t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()==idT)
	{
		if(!TableExists(t.getTokenName()))
		{
			pms.Format("Table %s does not exists",t.getTokenName());
			m_show(m_ln,pms,error);
			return -1;
		}
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=dotT)
		{
			m_show(m_ln,"Missing .* by table name",error);
			return -1;
		}
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=asterixT)
		{
			m_show(m_ln,"Missing .* by table name",error);
			return -1;
		}
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
	}
	if(t.getTokenType()==asterixT)
	{
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();

	}
	if(t.getTokenType()!=fromT)
	{
		m_show(m_ln,"Missing FROM",error);
		return -1;
	}
	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=idT)
	{
		m_show(m_ln,"Invalid table name",error);
		return -1;
	}

	if(!TableExists(t.getTokenName()))
	{
		pms.Format("Table %s does not exists",t.getTokenName());
		m_show(m_ln,pms,error);
		return -1;
	}
	
	CString table = t.getTokenName();

	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()==whereT)
	{
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		//WHERE clause id not checked
	}
	else
	{
		if(t.getTokenType()!=nilT && t.getTokenType()!=endfileT)
		{
			m_show(m_ln,"Semicolon missing behind the statement DELETE FROM",error);
			return -1;
		}
	}

	int ra = 0;
	try
	{
		m_dbs.Execute(m_cmd);
		ra = m_dbs.GetRecordsAffected();
	}
	catch (CDaoException* e)
	{
		m_show(m_ln,e->m_pErrorInfo->m_strDescription,error);
		e->Delete();
		return -1;
	}
	catch(...)
	{
		pms.Format("Error deleting data from %s",table);
		m_show(m_ln,pms,error);
		return -1;
	}

	if(m_verbose)
	{
		pms.Format("%i record(s) deleted from table %s\n",ra,table);
		m_show(m_ln,pms,info);
	}

	return 0;
}

int CSqlInterpreter::Drop()
{
	char msg[150];
	CString pms;

	m_lexan.NextToken();
	CToken t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()==tableT)
	{
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=idT)
		{
			m_show(m_ln,"Invalid table name",error);
			return -1;
		}

		if(!TableExists(t.getTokenName()))
		{
			pms.Format("Dropped table %s does not exists",t.getTokenName());
			m_show(m_ln,pms,warning);
			return 0;
		}

		sprintf(msg,"Table %s dropped",t.getTokenName());
	
	}
	else
	if(t.getTokenType()==indexT)
	{
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=idT)
		{
			m_show(m_ln,"Invalid index name",error);
			return -1;
		}

		if(!IndexExists(t.getTokenName()))
		{
			pms.Format("Dropped index %s does not exists",t.getTokenName());
			m_show(m_ln,pms,warning);
			return 0;
		}
		CString index = t.getTokenName();

		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=onT)
		{
			m_show(m_ln,"Missing ON in DROP INDEX statement",error);
			return -1;
		}

		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=idT)
		{
			m_show(m_ln,"Invalid table name",error);
			return -1;
		}

		if(!TableExists(t.getTokenName()))
		{
			pms.Format("Table %s does not exists",t.getTokenName());
			m_show(m_ln,pms,warning);
			return 0;
		}

		sprintf(msg,"Index %s on table %s dropped",index,t.getTokenName());
	
	}
	else
	if(t.getTokenType()==viewT)
	{
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=idT)
		{
			m_show(m_ln,"Invalid view name",error);
			return -1;
		}

		if(!QueryExists(t.getTokenName()))
		{
			pms.Format("Dropped view %s does not exists",t.getTokenName());
			m_show(m_ln,pms,warning);
			return 0;
		}

		try
		{
			m_dbs.DeleteQueryDef(t.getTokenName());
		}
		catch (CDaoException* e)
		{
			m_show(m_ln,e->m_pErrorInfo->m_strDescription,error);
			e->Delete();
			return -1;
		}
		catch(...)
		{
			pms.Format("Error dropping view %s",t.getTokenName());
			m_show(m_ln,pms,error);
			return -1;
		}

		if(m_verbose)
		{
			pms.Format("View %s dropped\n\n",t.getTokenName());
			m_show(m_ln,pms,info);
		}
		ReopenDatabase();
		return 0;
	
	}
	else
	{
		m_show(m_ln,"Missing TABLE,INDEX or VIEW in DROP statement",error);
		return -1;
	}

	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=nilT && t.getTokenType()!=endfileT)
	{
		m_show(m_ln,"Semicolon missing behind the statement DELETE FROM",error);
		return -1;
	}

	try
	{
		m_dbs.Execute(m_cmd);
	}
	catch (CDaoException* e)
	{
		m_show(m_ln,e->m_pErrorInfo->m_strDescription,error);
		e->Delete();
		return -1;
	}
	catch(...)
	{
		m_show(m_ln,"Error dropping table or index",error);
		return -1;
	}

	ReopenDatabase();
	if(m_verbose) m_show(m_ln,msg,info);
	return 0;
}

int CSqlInterpreter::Update()
{
	CString pms;

	m_lexan.NextToken();
	CToken t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=idT)
	{
		m_show(m_ln,"Invalid table name",error);
		return -1;
	}
	if(!TableExists(t.getTokenName()))
	{
		pms.Format("Table %s does not exists",t.getTokenName());
		m_show(m_ln,pms,error);
		return -1;
	}
	CString table = t.getTokenName();

	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=setT)
	{
		m_show(m_ln,"Missing SET in UPDATE statement",error);
		return -1;
	}
	for(;;)
	{
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=idT)
		{
			m_show(m_ln,"Field name missing",error);
			return -1;
		}
		CString field = t.getTokenName();

		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=ekvT)
		{
			pms.Format("= missing for field %s",field);
			m_show(m_ln,pms,error);
			return -1;
		}
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()!=intT && t.getTokenType()!=realT && t.getTokenType()!=stringT)
		{
			pms.Format("Value missing for field %s",field);
			m_show(m_ln,pms,error);
			return -1;
		}
		t = m_lexan.GetCurrentToken();
		m_lexan.NextToken();
		if(t.getTokenType()==whereT || t.getTokenType()==endfileT) break;
		if(t.getTokenType()!=commaT)
		{
			m_show(m_ln,"Comma missing",error);
			return -1;
		}
	}

	int ra=0;
	try
	{
		m_dbs.Execute(m_cmd);
		ra = m_dbs.GetRecordsAffected();
	}
	catch (CDaoException* e)
	{
		m_show(m_ln,e->m_pErrorInfo->m_strDescription,error);
		e->Delete();
		return -1;
	}
	catch(...)
	{
		pms.Format("Error updating table %s",table);
		m_show(m_ln,pms,error);
		return -1;
	}

	if(m_verbose)
	{
		pms.Format("%i record(s) updated in table %s",table);
		m_show(m_ln,pms,info);
	}
	
	return 0;
}

void CSqlInterpreter::AddMacro(const char * macro, CString value)
{
	m_macros.SetAt(macro,value);
}

int CSqlInterpreter::CreateView()
{
	CString pms;
	//name of the view
	CToken t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=idT)
	{
		pms.Format("Invalid view name - %s",t.getTokenName());
		m_show(m_ln,pms,error);
		return -1;
	}

	//check existence
	if(QueryExists(t.getTokenName()))
	{
		pms.Format("View %s already exists",t.getTokenName());
		m_show(m_ln,pms,error);
		return -1;
	}
	
	CString view = t.getTokenName();

	//AS must follow
	t = m_lexan.GetCurrentToken();
	m_lexan.NextToken();
	if(t.getTokenType()!=asT)
	{
		pms.Format("Missing AS in CREATE VIEW for view %s",t.getTokenName());
		m_show(m_ln,pms,error);
		return -1;
	}

	//stepping over CREATE VIEW .. AS
	char* cmd = stristr(m_cmd,"AS");
	cmd+=3;

	try
	{
		CDaoQueryDef qd(&m_dbs);
		qd.Create(view,cmd);
		qd.Append();
	}
	catch (CDaoException* e)
	{
		m_show(m_ln,e->m_pErrorInfo->m_strDescription,error);
		e->Delete();
		return -1;
	}
	catch(...)
	{
		pms.Format("Error creating view %s",view);
		m_show(m_ln,pms,error);
		return -1;
	}

	ReopenDatabase();
	if(m_verbose)
	{
		pms.Format("View %s created",view);
		m_show(m_ln,pms,info);
	}
	return 0;
}

void CSqlInterpreter::ExpandMacros(char* buf, int buflen)
{
	if(m_macros.GetCount())
	{
		CStringEx cmd(buf);
		CString macro,value;

		POSITION pos = m_macros.GetStartPosition();
		while(pos)
		{
			m_macros.GetNextAssoc(pos,macro,value);
			macro = '%'+macro+'%';
			cmd.FindReplace(macro,value);
		}

		strncpy(buf,cmd,buflen);
	}
}

int CSqlInterpreter::Select()
{
	CString msg;

	try
	{
		CDaoRecordset rst(&m_dbs);
		rst.Open(AFX_DAO_USE_DEFAULT_TYPE,m_cmd);
		int colcnt = rst.GetFieldCount();

		for(int ix=0; ix<colcnt; ix++)
		{
			CDaoFieldInfo fi;
			rst.GetFieldInfo(ix,fi);
			if(ix) msg+=", ";
			msg += fi.m_strName;
		}
		m_show(0,msg,list);
		CString pms('-',msg.GetLength());
		m_show(0,pms,list);

		while(!rst.IsEOF())
		{
			msg = "";
			for(int ix=0; ix<colcnt; ix++)
			{
				if(ix) msg+=", ";
				msg += VarToStr(rst.GetFieldValue(ix));
			}
			m_show(0,msg,list);
			rst.MoveNext();
		}
		rst.Close();
		m_show(0,"",list);
	}
	catch(CDaoException *e)
	{
		m_show(m_ln,e->m_pErrorInfo->m_strDescription,error);
		e->Delete();
		return -1;
	}
	catch(...)
	{
		m_show(m_ln,"Error executing SELECT command",error);
		return -1;
	}
	return 0;
}
