//////////////////////////////////////////////////////////////////////
// JetSQL - MS Jet SQL script interpreter (c) TECHSYS HW a SW s.r.o.
//////////////////////////////////////////////////////////////////////
// Module: Interpreter.h
// Author: Pavel Klocek
// Desc:   Definition of class CSqlInterpreter 
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_INTERPRETER_H__B539FFE1_A21C_11D2_B6BB_BDA20BB5300A__INCLUDED_)
#define AFX_INTERPRETER_H__B539FFE1_A21C_11D2_B6BB_BDA20BB5300A__INCLUDED_

#include "LexicalAnalyser.h"

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

//container for macros
typedef CMap<CString, LPCSTR, CString, CString&> MACROS;

//callback function for displaying messages from interpreter
//types:
//  echoln - lines readed from input if the echo mode is on
//  info - information about execution results if the verbose mode is on
//  list - lines of SELECT command ouutputs
//  warnign
//  error - description of an error in SQL command
//  fatalerror - description of an error whitch prevents the execution
typedef enum{echoln,info,list,warning,error,fatalerror} SQLSHOWTYPE;
typedef void SQLSHOWFUNC(int line, const char *txt, SQLSHOWTYPE type);

//////////////////////////////////////////////////////////////////////
// class CSqlInterpreter
//--------------------------------------------------------------------
// This class does the work
//////////////////////////////////////////////////////////////////////
class CSqlInterpreter  
{
public:
//////////////////////////////////////////////////////////////////////
// Initialization

	//constructor initializes lexical analyser for SQL
	CSqlInterpreter();
	virtual ~CSqlInterpreter();
	
	//Adds a macro-value pair
	void AddMacro(const char* macro, CString value);

//////////////////////////////////////////////////////////////////////
// Execution

	//Opens the database, parses the sql file and executes its commands.
	//
	//Returns number of errors during execution, -1 for fatal error
	//
	//dbfile ... path to database file (*.mdb)
	//sqlfile ... path to sql script file
	//show ... pointer to callback function receiving messages
	//create ... if true, new database file is created (old file is rewrited)
	//verbose ... if true, messages about succesfully completed command are displayed
	//echo ... if true, executed commands are reported.
	int Go(CString dbfile, CString sqlfile, SQLSHOWFUNC& show, bool create = false, bool verbose = false, bool echo = false);

protected:

//////////////////////////////////////////////////////////////////////
// Parsing

	//Executes command referenced by m_cmd
	int Execute();
	
	//Parses and executes UPDATE command
	int Update();
	//Parses and executes DROP command
	int Drop();
	//Parses and executes DELETE command
	int Delete();
	//Parses and executes INSERT INTO command
	int InsertInto();
	//Parses and executes CREATE INDEX command
	int CreateIndex();
	//Parses and executes CREATE TABLE command
	int CreateTable();
	//Parses and executes CREATE VIEW command
	int CreateView();
	//Lists results of SELECT command
	int Select();

	//Determines, if an index exists in the database
	bool IndexExists(const char* name);
	//Determines, if a table exists in the database
	bool TableExists(const char* name);
	//Determines, if a query exists in the database
	bool QueryExists(const char* name);

//////////////////////////////////////////////////////////////////////
// Helper methods

	//expand macros in buffer
	void ExpandMacros(char* buf, int buflen);
	//Opens or creates the database file
	int OpenDatabase(CString dbfile, bool create);
	//Closes and opens the database (needed to refresh DAO data structures
	//after executin DDL SQL commands like CREATE TABLE)
	int ReopenDatabase();

//////////////////////////////////////////////////////////////////////
// Attributes

	CLexicalAnalyser m_lexan;	//lexical analyser
	CString m_dbfile;			//filename of the current database
	CDaoDatabase m_dbs;			//current database
	MACROS m_macros;			//macros for text substitition
	int m_ln;					//current line
	char* m_cmd;				//current cmd
	bool m_verbose;				//verbose mode
	SQLSHOWFUNC* m_show;		//current callback function

};

#endif // !defined(AFX_INTERPRETER_H__B539FFE1_A21C_11D2_B6BB_BDA20BB5300A__INCLUDED_)
